#!/usr/bin/python

import math
import numpy
import psycopg2
import random
import re
import sys

HEAD_PORT = 5433
PATCH_PORT = 5462
TEST_DATA_FILE = '/tmp/mcv-test-data'

conn1 = psycopg2.connect('port=%d dbname=pgdevel user=pgdevel' % HEAD_PORT)
conn2 = psycopg2.connect('port=%d dbname=pgdevel user=pgdevel' % PATCH_PORT)
conn1.set_session(autocommit = True)
conn2.set_session(autocommit = True)
cur1 = conn1.cursor()
cur2 = conn2.cursor()

def recreate_table():
    for cur in [cur1, cur2]:
        cur.execute('DROP TABLE IF EXISTS mcv_test')
        cur.execute('CREATE TABLE mcv_test(x int)')

def load_data():
    for cur in [cur1, cur2]:
        cur.execute('COPY mcv_test FROM \'%s\'' % TEST_DATA_FILE)
        cur.execute('CREATE INDEX mcv_test_x_idx ON mcv_test(x)')

def test_stats(cur, vals):
    total_err = 0
    max_err = 0
    for x in vals:
        cur.execute('EXPLAIN ANALYSE'\
                    ' SELECT * FROM mcv_test WHERE x = %s', (x,))
        plan = cur.fetchone()[0]
        match = re.search('.*rows=([0-9]*).*rows=([0-9]*)', plan)
        estimate = float(match.group(1))
        actual = float(match.group(2))
        diff = estimate - actual
        err = 100 * abs(diff) / (actual + 10)
        total_err += err
        max_err = max(max_err, err)
    return (total_err, max_err)

def print_results(results):
    cells = []
    for result in results:
        cells.append([result[0]] + [str(x) for x in result[1]])

    widths = [0 for x in cells[0]]
    for row in cells:
        for i in range(0, len(row)):
            widths[i] = max(widths[i], len(row[i]))

    rownum = 0
    for row in cells:
        rownum += 1

        outputs = ['%*s' % (widths[0], row[0])]
        for i in range(1, len(row)):
            outputs.append('%*s' % (-widths[i], row[i]))
        print ' | '.join(outputs)

        if rownum == 2:
            seps = ['-'*x for x in widths]
            print '-+-'.join(seps)

    print ''
    sys.stdout.flush()

def test_analyse():
    branches = []
    targets = []
    num_mcvs = []
    mcv_lists = []
    counts = []
    dcounts = []
    min_vals = []
    max_vals = []
    avg_mcv_errors = []
    avg_non_mcv_errors = []
    avg_errors = []
    max_mcv_errors = []
    max_non_mcv_errors = []
    max_errors = []

    for (cur, target, branch) in [ (cur1, 100, 'HEAD'),
                                   (cur2, 100, 'Patch'),
                                   (cur1, 1000, 'HEAD'),
                                   (cur2, 1000, 'Patch')]:
        branches.append(branch)
        targets.append('(tgt=%d)' % target)

        cur.execute('ALTER TABLE mcv_test '\
                    'ALTER COLUMN x SET STATISTICS %d' % target)
        cur.execute('ANALYSE mcv_test')
        cur.execute('SELECT coalesce(cardinality(most_common_vals), 0)'\
                    '  FROM pg_stats WHERE tablename = \'mcv_test\'')
        num_mcvs.append(cur.fetchone()[0])

        cur.execute('SELECT unnest('\
                    '         string_to_array('\
                    '           array_to_string('\
                    '             most_common_vals, \',\'), \',\'))::int'\
                    '  FROM pg_stats WHERE tablename = \'mcv_test\''\
                    ' ORDER BY 1')
        mcv_list = [x[0] for x in cur.fetchall()]
        mcv_set = set(mcv_list)
        mcv_lists.append(mcv_list)

        cur.execute('SELECT count(*), count(distinct x), min(x), max(x)'\
                    '  FROM mcv_test')
        (count, dcount, min_val, max_val) = cur.fetchone()
        counts.append(count)
        dcounts.append(dcount)
        min_vals.append(min_val)
        max_vals.append(max_val)

        tot = max_val - min_val
        cnt = min(tot, 1000 + len(mcv_list))
        non_mcv_list = []
        for i in range(0, cnt+1):
            x = min_val + int(i*float(tot)/float(cnt))
            if not x in mcv_set: non_mcv_list.append(x)

        (total_mcv_err, max_mcv_err) = test_stats(cur, mcv_list)
        (total_non_mcv_err, max_non_mcv_err) = test_stats(cur, non_mcv_list)
        total_err = total_mcv_err + total_non_mcv_err
        max_err = max(max_mcv_err, max_non_mcv_err)

        if len(mcv_list) > 0:
            avg_mcv_err = total_mcv_err / len(mcv_list)
            avg_mcv_errors.append('%.2f%%' % avg_mcv_err)
            max_mcv_errors.append('%.2f%%' % max_mcv_err)
        else:
            avg_mcv_errors.append('-')
            max_mcv_errors.append('-')

        if len(non_mcv_list) > 0:
            avg_non_mcv_err = total_non_mcv_err / len(non_mcv_list)
            avg_non_mcv_errors.append('%.2f%%' % avg_non_mcv_err)
            max_non_mcv_errors.append('%.2f%%' % max_non_mcv_err)
        else:
            avg_non_mcv_errors.append('-')
            max_non_mcv_errors.append('-')

        avg_err = total_err / (len(mcv_list) + len(non_mcv_list))
        avg_errors.append('%.2f%%' % avg_err)
        max_errors.append('%.2f%%' % max_err)

    results = [ ['', branches],
                ['', targets],
                ['Table size:', counts],
                ['Distinct values:', dcounts],
                ['Minimum value:', min_vals],
                ['Maximum value:', max_vals],
                ['Number of MCVs:', num_mcvs],
                ['Avg. MCV error:', avg_mcv_errors],
                ['Max. MCV error:', max_mcv_errors],
                ['Avg. non-MCV error:', avg_non_mcv_errors],
                ['Max. non-MCV error:', max_non_mcv_errors],
                ['Overall avg. error:', avg_errors],
                ['Overall max. error:', max_errors] ]
    print_results(results)

def print_title(title):
    print title
    print '='*len(title)
    print ''
    sys.stdout.flush()

def uniform_correlated(fd, N, Nd):
    print_title('Correlated uniform distribution (N=%d, Nd=%d)' % (N, Nd))
    x = 0
    while x < N:
        val = int(x*float(Nd)/N)
        print >>fd, val
        x += 1

def uniform_random(fd, N, Nd):
    print_title('Random uniform distribution (N=%d, Nd=%d)' % (N, Nd))
    vals = range(0, N)
    random.shuffle(vals)
    x = 0
    while x < N:
        val = int(vals[x]*float(Nd)/N)
        print >>fd, val
        x += 1

def normal(fd, N, sd):
    print_title('Normal distribution (N=%d, sd=%d)' % (N, sd))
    for val in numpy.random.normal(0, sd, N):
        print >>fd, int(round(val))

def binomial(fd, N, n, p):
    print_title('Binomial distribution (N=%d, n=%d, p=%f)' % (N, n, p))
    for val in numpy.random.binomial(n, p, N):
        print >>fd, int(round(val))

def exponential(fd, N, sd):
    print_title('Exponential distribution (N=%d, sd=%f (beta))' % (N, sd))
    for val in numpy.random.exponential(sd, N):
        print >>fd, int(round(val))

def laplace(fd, N, b):
    print_title('Laplace distribution (N=%d, b=%f (lambda))' % (N, b))
    for val in numpy.random.laplace(0, b, N):
        print >>fd, int(round(val))

def pareto(fd, N, a):
    print_title('Pareto distribution (N=%d, a=%f)' % (N, a))
    for val in numpy.random.pareto(a, N):
        print >>fd, int(round(val))

def multimodal(fd, N, peaks, sep, sd):
    print_title('Multimodal normal mix (N=%d, peaks=%d, sep=%d, sd=%d)'
                % (N, peaks, sep, sd))
    for x in range(0, N):
        mean = random.randrange(0, peaks)*sep
        print >>fd, int(round(numpy.random.normal(mean, sd)))

def test(dist_fn, *args):
    recreate_table()
    with open(TEST_DATA_FILE, 'w') as fd:
        dist_fn(fd, *args)
    load_data()
    test_analyse()

def uniform_correlated_tests():
    test(uniform_correlated, 30000, 100)
    test(uniform_correlated, 30000, 150)

    test(uniform_correlated, 40000, 100)
    test(uniform_correlated, 40000, 150)
    test(uniform_correlated, 40000, 200)
    test(uniform_correlated, 40000, 500)
    test(uniform_correlated, 40000, 1000)
    test(uniform_correlated, 40000, 2000)
    test(uniform_correlated, 40000, 3000)
    test(uniform_correlated, 40000, 4000)
    test(uniform_correlated, 40000, 5000)
    test(uniform_correlated, 40000, 10000)
    test(uniform_correlated, 40000, 20000)

    test(uniform_correlated, 50000, 100)
    test(uniform_correlated, 50000, 150)
    test(uniform_correlated, 50000, 200)
    test(uniform_correlated, 50000, 500)
    test(uniform_correlated, 50000, 1000)
    test(uniform_correlated, 50000, 2000)
    test(uniform_correlated, 50000, 3000)
    test(uniform_correlated, 50000, 6000)
    test(uniform_correlated, 50000, 12000)
    test(uniform_correlated, 50000, 25000)

    test(uniform_correlated, 60000, 100)
    test(uniform_correlated, 60000, 150)
    test(uniform_correlated, 60000, 200)
    test(uniform_correlated, 60000, 500)
    test(uniform_correlated, 60000, 1000)
    test(uniform_correlated, 60000, 2000)
    test(uniform_correlated, 60000, 3000)
    test(uniform_correlated, 60000, 5000)
    test(uniform_correlated, 60000, 7000)
    test(uniform_correlated, 60000, 15000)
    test(uniform_correlated, 60000, 30000)

    test(uniform_correlated, 100000, 100)
    test(uniform_correlated, 100000, 150)
    test(uniform_correlated, 100000, 200)
    test(uniform_correlated, 100000, 300)
    test(uniform_correlated, 100000, 400)
    test(uniform_correlated, 100000, 500)
    test(uniform_correlated, 100000, 1000)
    test(uniform_correlated, 100000, 2000)
    test(uniform_correlated, 100000, 3000)
    test(uniform_correlated, 100000, 4000)
    test(uniform_correlated, 100000, 5000)
    test(uniform_correlated, 100000, 10000)
    test(uniform_correlated, 100000, 20000)
    test(uniform_correlated, 100000, 50000)

    test(uniform_correlated, 1000000, 100)
    test(uniform_correlated, 1000000, 200)
    test(uniform_correlated, 1000000, 500)
    test(uniform_correlated, 1000000, 750)
    test(uniform_correlated, 1000000, 1000)
    test(uniform_correlated, 1000000, 1500)
    test(uniform_correlated, 1000000, 2000)
    test(uniform_correlated, 1000000, 3000)
    test(uniform_correlated, 1000000, 4000)
    test(uniform_correlated, 1000000, 5000)
    test(uniform_correlated, 1000000, 10000)
    test(uniform_correlated, 1000000, 100000)
    test(uniform_correlated, 1000000, 500000)

def uniform_random_tests():
    test(uniform_random, 30000, 100)
    test(uniform_random, 30000, 150)

    test(uniform_random, 40000, 100)
    test(uniform_random, 40000, 150)
    test(uniform_random, 40000, 200)
    test(uniform_random, 40000, 500)
    test(uniform_random, 40000, 1000)
    test(uniform_random, 40000, 2000)
    test(uniform_random, 40000, 3000)
    test(uniform_random, 40000, 4000)
    test(uniform_random, 40000, 5000)
    test(uniform_random, 40000, 10000)
    test(uniform_random, 40000, 20000)

    test(uniform_random, 50000, 100)
    test(uniform_random, 50000, 150)
    test(uniform_random, 50000, 200)
    test(uniform_random, 50000, 500)
    test(uniform_random, 50000, 1000)
    test(uniform_random, 50000, 2000)
    test(uniform_random, 50000, 3000)
    test(uniform_random, 50000, 6000)
    test(uniform_random, 50000, 12000)
    test(uniform_random, 50000, 25000)

    test(uniform_random, 60000, 100)
    test(uniform_random, 60000, 150)
    test(uniform_random, 60000, 200)
    test(uniform_random, 60000, 500)
    test(uniform_random, 60000, 1000)
    test(uniform_random, 60000, 2000)
    test(uniform_random, 60000, 3000)
    test(uniform_random, 60000, 5000)
    test(uniform_random, 60000, 7000)
    test(uniform_random, 60000, 15000)
    test(uniform_random, 60000, 30000)

    test(uniform_random, 100000, 100)
    test(uniform_random, 100000, 150)
    test(uniform_random, 100000, 200)
    test(uniform_random, 100000, 300)
    test(uniform_random, 100000, 400)
    test(uniform_random, 100000, 500)
    test(uniform_random, 100000, 1000)
    test(uniform_random, 100000, 2000)
    test(uniform_random, 100000, 3000)
    test(uniform_random, 100000, 4000)
    test(uniform_random, 100000, 5000)
    test(uniform_random, 100000, 10000)
    test(uniform_random, 100000, 20000)
    test(uniform_random, 100000, 50000)

    test(uniform_random, 1000000, 100)
    test(uniform_random, 1000000, 200)
    test(uniform_random, 1000000, 500)
    test(uniform_random, 1000000, 750)
    test(uniform_random, 1000000, 1000)
    test(uniform_random, 1000000, 1500)
    test(uniform_random, 1000000, 2000)
    test(uniform_random, 1000000, 3000)
    test(uniform_random, 1000000, 4000)
    test(uniform_random, 1000000, 5000)
    test(uniform_random, 1000000, 10000)
    test(uniform_random, 1000000, 100000)
    test(uniform_random, 1000000, 500000)

def normal_tests():
    test(normal, 500000, 1)
    test(normal, 500000, 2)
    test(normal, 500000, 3)
    test(normal, 500000, 4)
    test(normal, 500000, 5)
    test(normal, 500000, 6)
    test(normal, 500000, 6)
    test(normal, 500000, 7)
    test(normal, 500000, 8)
    test(normal, 500000, 9)
    test(normal, 500000, 10)
    test(normal, 500000, 15)
    test(normal, 500000, 20)
    test(normal, 500000, 30)
    test(normal, 500000, 40)
    test(normal, 500000, 50)
    test(normal, 500000, 75)
    test(normal, 500000, 100)
    test(normal, 500000, 150)
    test(normal, 500000, 200)
    test(normal, 500000, 300)
    test(normal, 500000, 400)
    test(normal, 500000, 500)
    test(normal, 500000, 600)
    test(normal, 500000, 800)
    test(normal, 500000, 1000)

def binomial_tests():
    test(binomial, 500000, 100, 0.05)
    test(binomial, 500000, 100, 0.1)
    test(binomial, 500000, 100, 0.2)
    test(binomial, 500000, 100, 0.3)
    test(binomial, 500000, 100, 0.4)
    test(binomial, 500000, 100, 0.5)

    test(binomial, 500000, 1000, 0.1)
    test(binomial, 500000, 1000, 0.2)
    test(binomial, 500000, 1000, 0.3)
    test(binomial, 500000, 1000, 0.4)
    test(binomial, 500000, 1000, 0.5)

    test(binomial, 500000, 2000, 0.5)
    test(binomial, 500000, 3000, 0.5)
    test(binomial, 500000, 4000, 0.5)
    test(binomial, 500000, 5000, 0.5)

def exponential_tests():
    test(exponential, 500000, 0.25)
    test(exponential, 500000, 0.5)
    test(exponential, 500000, 1)
    test(exponential, 500000, 2)
    test(exponential, 500000, 3)
    test(exponential, 500000, 4)
    test(exponential, 500000, 5)
    test(exponential, 500000, 10)
    test(exponential, 500000, 20)
    test(exponential, 500000, 30)

def laplace_tests():
    test(laplace, 500000, 0.1)
    test(laplace, 500000, 0.25)
    test(laplace, 500000, 0.5)
    test(laplace, 500000, 1)
    test(laplace, 500000, 2)
    test(laplace, 500000, 3)
    test(laplace, 500000, 4)
    test(laplace, 500000, 5)
    test(laplace, 500000, 10)
    test(laplace, 500000, 25)
    test(laplace, 500000, 50)
    test(laplace, 500000, 100)

def pareto_tests():
    test(pareto, 500000, 1)
    test(pareto, 500000, 2)
    test(pareto, 500000, 3)
    test(pareto, 500000, 4)
    test(pareto, 500000, 5)

def multimodal_tests():
    test(multimodal, 500000, 2, 100, 10)
    test(multimodal, 500000, 3, 50, 10)
    test(multimodal, 500000, 5, 30, 10)
    test(multimodal, 1000000, 10, 20, 10)
    test(multimodal, 2000000, 20, 20, 15)

uniform_correlated_tests()
uniform_random_tests()
normal_tests()
binomial_tests()
exponential_tests()
laplace_tests()
pareto_tests()
multimodal_tests()
